@@ -1,10 +1,11 @@
 import json
-from typing import Union, List, Dict
+import os.path
+from typing import Union, List, Dict, Optional
 
 from janis_assistant.data.enums.taskstatus import TaskStatus
 from janis_assistant.data.models.run import RunModel
 from janis_assistant.data.models.workflowjob import RunJobModel
-from janis_assistant.utils.dateutil import DateUtil
+from janis_assistant.utils.dateutils import DateUtil
 from janis_core.utils.logger import Logger
 
 
@@ -138,7 +139,9 @@ class CromwellMetadata:
         return CromwellMetadata.unwrap_caused_by(self.meta["failures"])
 
     @staticmethod
-    def unwrap_caused_by(d) -> str:
+    def unwrap_caused_by(d: Union[List, Dict]) -> Optional[str]:
+        if d is None:
+            return None
         if isinstance(d, list):
             return ", ".join(CromwellMetadata.unwrap_caused_by(x) for x in d)
 
@@ -196,6 +199,8 @@ class CromwellMetadata:
             start = start or min(s.start for s in subjobs)
             finish = finish
 
+        callroot = call.get("callRoot")
+
         return RunJobModel(
             id_=jid,
             submission_id=None,
@@ -221,7 +226,12 @@ class CromwellMetadata:
             analysis=None,
             memory=None,
             cpu=None,
-            script=None,
+            error=CromwellMetadata.unwrap_caused_by(call.get("failures")),
+            returncode=call.get("returnCode"),
+            workdir=call.get("callRoot"),
+            script=("file:/" + os.path.join(callroot, "execution/script"))
+            if callroot
+            else None,
         )
 
     @classmethod
